<?php if ( ! defined( 'ABSPATH' ) ) exit; 
/**
 * Plugin Name: FooSales
 * Description: FooSales is a secure point of sale (POS) system for WooCommerce that turns any computer, iPad or Android tablet into a cash register and makes it easy to sell your products and run your business from anywhere in the world. The FooSales WordPress plugin connects your online store to the FooSales apps.
 * Version: 1.10.6
 * Author: FooSales
 * Author URI: https://www.foosales.com/
 * Developer: FooSales
 * Developer URI: https://www.foosales.com/
 * Text Domain: foosales
 *
 * Copyright: 2009-2017 Grenade Technologies.
 * License: GNU General Public License v3.0
 * License URI: http://www.gnu.org/licenses/gpl-3.0.html
 */

//include config
require('config.php');

class FooSales {
    
    private $Config;
    private $WooHelper;
    private $XMLRPCHelper;
    
    public function __construct() {
        
        add_action( 'init', array( $this, 'plugin_init' ) );
        add_action( 'admin_notices', array( $this, 'check_woocommerce' ) );
        add_action( 'activated_plugin', array( $this, 'foosales_activation_redirect' ) );
        add_action( 'admin_init', array(&$this, 'assign_admin_caps' ) );
        add_action( 'admin_init', array( $this, 'foosales_register_styles' ) );
        add_filter( 'parse_query', array( $this, 'foosales_filter_order_results' ) );
        add_action( 'restrict_manage_posts', array( $this, 'foosales_filter_orders') );
        add_action( 'manage_shop_order_posts_custom_column', array( $this, 'foosales_order_column_content') );
        add_action( 'woocommerce_admin_order_data_after_order_details', array( $this, 'foosales_order_meta_general') );
        add_action('admin_menu', array($this, 'add_woocommerce_submenu'));

        add_filter( 'woocommerce_get_sections_advanced', array($this, 'foosales_add_section') );
        add_filter( 'woocommerce_get_settings_advanced', array($this, 'foosales_all_settings'), 10, 2 );
        add_filter( 'plugin_action_links_' . plugin_basename(__FILE__), array( $this, 'add_action_links' ) );        
		add_filter( 'woocommerce_rest_prepare_product_object', array($this, 'prepare_product_images'), 10, 3);
		add_filter( 'woocommerce_rest_prepare_product_variation_object', array($this, 'prepare_product_variation_image'), 10, 3);
        add_filter( 'manage_edit-shop_order_columns', array($this, 'foosales_order_column') , 20 ); 
        
        register_deactivation_hook( __FILE__, array( &$this, 'remove_admin_caps' ) );
    }

    /**
     * Register plugin styles.
     * 
     */
    public function foosales_register_styles() {

        wp_enqueue_style('foosales-scripts',  $this->Config->stylesPath . 'style.css', array(), '1.0.'.rand(5, 15)); 

    } 
	
	function prepare_product_images($response, $post, $request) {

		global $_wp_additional_image_sizes;

		if ( empty($response->data) ) {
			return $response;
		}

		foreach ( $response->data['images'] as $key => $image ) {
			$image_info = wp_get_attachment_image_src($image['id'], 'thumbnail');
			$response->data['images'][$key]['thumbnail'] = $image_info[0];
		}
		
		return $response;
	}

    function prepare_product_variation_image($response, $post, $request) {

		if ( empty($response->data) ) {
			return $response;
		}
		
		$image_info = wp_get_attachment_image_src($response->data['image']['id'], 'thumbnail');
		$response->data['image']['thumbnail'] = $image_info[0];
		
		$response->data['price_html'] = wc_get_product($response->data['id'])->get_price_html();
		
		return $response;
	}
    
   function foosales_add_section( $sections ) {

        $sections['foosales'] = __( 'FooSales', 'foosales' );

        return $sections;
    }

    function foosales_all_settings( $settings, $current_section ) {

	    if ( $current_section == 'foosales' ) {

            $foosales_settings = array();

            $foosales_settings[] = array(
                'name' => __('FooSales Settings', 'foosales'),
                'type' => 'title',
                'id' => 'wc_settings_foosales_settings_title',
            );

            $foosales_settings[] = array('type' => 'sectionstart', 'id' => 'foosales_settings_products');

            $foosales_settings[] = array(
                'name' => __('Products', 'foosales'),
                'type' => 'title',
                'id' => 'wc_settings_foosales_products_title',
            );

            // Get all product categories
            $taxonomy = 'product_cat';
            $orderby = 'name';
            $show_count = 1; // 1 for yes, 0 for no
            $pad_counts = 0; // 1 for yes, 0 for no
            $hierarchical = 1; // 1 for yes, 0 for no
            $title = '';
            $empty = 0;

            $args = array(
                'taxonomy' => $taxonomy,
                'orderby' => $orderby,
                'show_count' => $show_count,
                'pad_counts' => $pad_counts,
                'hierarchical' => $hierarchical,
                'title_li' => $title,
                'hide_empty' => $empty,
            );

            $all_categories = get_categories($args);

            $cat_options = array();

            foreach ($all_categories as $cat) {

                if ($cat->category_parent == 0) {

                    $category_id = $cat->term_id;

                    $cat_options[$cat->term_id] = $cat->name;

                    $args2 = array(
                        'taxonomy' => $taxonomy,
                        'child_of' => 0,
                        'parent' => $category_id,
                        'orderby' => $orderby,
                        'show_count' => $show_count,
                        'pad_counts' => $pad_counts,
                        'hierarchical' => $hierarchical,
                        'title_li' => $title,
                        'hide_empty' => $empty,
                    );

                    $sub_cats = get_categories($args2);

                    if ( $sub_cats ) {
                        foreach ( $sub_cats as $sub_category ) {
                            $cat_options[$sub_category->term_id] = '   - ' . $sub_category->name;
                        }
                    }
                }
            }

            $foosales_settings[] = array(
                'name' => __('Products to display', 'foosales'),
                'type' => 'radio',
                'id' => 'globalFooSalesProductsToDisplay',
                'class' => 'text',
                'default' => 'all',
                'options' => array(
                    'all' => __('Show all products', 'foosales'),
                    'cat' => __('Only show products from these specific categories:', 'foosales'),
                ),
            );

            $foosales_settings[] = array(
                'type' => 'multiselect',
                'id' => 'globalFooSalesProductCategories',
                'class' => 'text',
                'options' => $cat_options,
                'css' => (count($cat_options) < 10) ? 'height:100%' : 'height:200px',
            );

            $foosales_settings[] = array('type' => 'sectionend', 'id' => 'foosales_settings_products');
            $foosales_settings[] = array('type' => 'sectionstart', 'id' => 'foosales_settings_receipts');

            $foosales_settings[] = array(
                'name' => __('Receipts/Tax Invoices', 'foosales'),
                'type' => 'title',
                'desc' => __('These details are optional and will be displayed on the receipts/tax invoices printed from orders generated by FooSales.', 'foosales'),
                'id' => 'wc_settings_foosales_receipts_title',
            );

            $foosales_settings[] = array(
                'name' => __('Store logo URL', 'foosales'),
                'type' => 'text',
                'id' => 'globalFooSalesStoreLogoURL',
                'desc' => __('The URL that links to a black and white version of your store logo.', 'foosales'),
                'class' => 'text',
            );

            $foosales_settings[] = array(
                'name' => __('Store name', 'foosales'),
                'type' => 'text',
                'id' => 'globalFooSalesStoreName',
                'class' => 'text',
            );

            $foosales_settings[] = array(
                'name' => __('Header content', 'foosales'),
                'type' => 'textarea',
                'id' => 'globalFooSalesHeaderContent',
                'class' => 'text',
            );

            $foosales_settings[] = array(
                'name' => __('Footer content', 'foosales'),
                'type' => 'textarea',
                'id' => 'globalFooSalesFooterContent',
                'class' => 'text',
            );

            $foosales_settings[] = array('type' => 'sectionend', 'id' => 'foosales_settings_receipts');
            $foosales_settings[] = array('type' => 'sectionend', 'id' => 'foosales');

            return $foosales_settings;
        } else {
            return $settings;
        }
    }
    
    /**
     *  Initialize plugin and helpers.
     * 
     */
    public function plugin_init() {
        
        //Main config
        $this->Config = new FooSales_Config();
        
        //WooHelper
        require_once($this->Config->classPath.'woohelper.php');
        $this->WooHelper = new FooSales_Woo_Helper($this->Config);
        
        //XMLRPCHelper
        require_once($this->Config->classPath.'xmlrpchelper.php');
        $this->XMLRPCHelper = new FooSales_XMLRPC_Helper($this->Config);
    }
    
    /**
     * Checks if WooCommerce is active.
     * 
     */
    public function check_woocommerce() {

        if ( !class_exists( 'WooCommerce' ) ) {
            $this->output_notices(array(__( 'WooCommerce is not active. Please install and activate it.', 'foosales' )));
        }
    }
    
    /**
     * Outputs notices to screen.
     * 
     * @param array $notices
     */
    private function output_notices($notices) {

        foreach ($notices as $notice) {
           echo "<div class='updated'><p>$notice</p></div>";
        }
    }
    
    public function assign_admin_caps() {
        
        $role = get_role( 'administrator' );
        
        $role->add_cap('publish_foosales');
    }
    
    public function remove_admin_caps() {
        
        $delete_caps = array(
            'publish_foosales'
        );
        
        global $wp_roles;

	    foreach ($delete_caps as $cap) {

            foreach (array_keys($wp_roles->roles) as $role) {

                $wp_roles->remove_cap($role, $cap);
            }        
	    }
    }

    /**
     * Adds settings link to plugin listing
     * 
     */
    public function add_action_links ( $links ) {

        $linkSettings = '<a href="' . admin_url( 'admin.php?page=wc-settings&tab=advanced&section=foosales' ) . '">Settings</a>';
        array_unshift($links, $linkSettings); 
        
        $linkIntroduction = '<a href="admin.php?page=foosales-introduction">'.__( 'Introduction', 'foosales' ).'</a>'; 
        array_unshift($links, $linkIntroduction); 
        
        return $links;        
 
    }

    public function foosales_activation_redirect( $plugin ) {

        if ( $plugin == plugin_basename( __FILE__ ) ) {
            exit( wp_redirect( admin_url( 'admin.php?page=foosales-introduction' ) ) ); 
        }
    }        

    /**
     * Filter WooCommerce orders listing based on FooSales filter selection
     * 
     */
    public function foosales_filter_order_results( $query ) {
        global $pagenow;
        $foosales_filter=''; 
        if ( is_admin() && $pagenow=='edit.php' && isset($_GET['foosales_filter']) && $_GET['foosales_filter'] != '') {
            $foosales_filter = esc_attr($_GET['foosales_filter']);
            switch ($foosales_filter) {
                case "pos-all":
                    // All FooSales payments
                    $query->query_vars['meta_key'] = '_foosales_order_source';
                    $query->query_vars['meta_value'] = 'foosales_app';
                    break;
                case "pos-card":
                    $query->query_vars['meta_key'] = '_foosales_payment_method';
                    $query->query_vars['meta_value'] = 'foosales_card';
                    break;
                case "pos-cash":
                    $query->query_vars['meta_key'] = '_foosales_payment_method';
                    $query->query_vars['meta_value'] = 'foosales_cash';
                    break;
                case "pos-none":
                    // Online orders only 
                    $query->query_vars['meta_key'] = '_foosales_order_source';
                    $query->query_vars['meta_compare'] = 'NOT EXISTS';
                    $query->query_vars['meta_value'] = 'foosales_app';
                    break;
            }     
        }
    }

    /**
     * Adds FooSales drop down filter selection to the WooCommerce orders listing
     * 
     */
    public function foosales_filter_orders() {
        $foosales_filter=''; 
        if ( isset($_GET['foosales_filter']) && $_GET['foosales_filter'] != '') {
            $foosales_filter = $_GET['foosales_filter'];
        }
        global $wpdb;
        $sql = 'SELECT DISTINCT meta_key FROM '.$wpdb->postmeta.' ORDER BY 1';
        $fields = $wpdb->get_results($sql, ARRAY_N);
        ?>
        <select name="foosales_filter">
            <option value=""><?php _e('All Orders', 'foosales'); ?></option>
            <option value="pos-all" <?php if($foosales_filter =='pos-all') { echo 'selected';  } ?> ><?php _e('POS Orders', 'foosales'); ?></option>
            <option value="pos-card" <?php if($foosales_filter =='pos-card') { echo 'selected';  } ?> ><?php _e('POS Card Payments', 'foosales'); ?></option>
            <option value="pos-cash" <?php if($foosales_filter =='pos-cash') { echo 'selected';  } ?> ><?php _e('POS Cash Payments', 'foosales'); ?></option>
            <option value="pos-none" <?php if($foosales_filter =='pos-none') { echo 'selected';  } ?>><?php _e('Online Orders', 'foosales'); ?></option>
        </select>
        <?php
    }    

    /**
     * Add POS column to WooCommerce orders listing
     * 
     */
    public function foosales_order_column( $columns ) {
        $columns['foosale_column_type'] = 'POS';
        return $columns;
    }

    public function foosales_order_column_content( $column ) {
        global $post;

        if ( 'foosale_column_type' === $column ) {

            $foosales_order_type =  get_post_meta( $post->ID , '_foosales_payment_method', true );
            switch ($foosales_order_type) {
                case "foosales_cash":
                    echo '<mark class="order-status foosale_type_cash"><span>'.__('Cash', 'foosales').'</span></mark>';
                    break;
                case "foosales_card":
                    echo '<mark class="order-status foosale_type_card"><span>'.__('Card', 'foosales').'</span></mark>';
                    break;
                default:
                    echo '<mark class="order-status foosale_type_online"><span>'.__('Online', 'foosales').'</span></mark>';
            }          
        }
    }    

    /**
     * Populate POS column on WooCommerce orders listing
     * 
     */
    public function foosales_order_meta_general( $order ) {

        $foosales_source = get_post_meta( $order->id, '_foosales_order_source', true );
        $foosales_payment_method = get_post_meta( $order->id, '_foosales_payment_method', true ); 
        $foosales_payment_method_pub = get_post_meta( $order->id, 'Order Payment Method', true ); 

        echo "<br class='clear' />";
        echo "<p><em>";
        echo __('Source:', 'foosales');
        if($foosales_source =='foosales_app' ) {
            echo __('POS', 'foosales');
        } else {
            echo __('Online', 'foosales');
        } 
        if($foosales_payment_method =='Card' or $foosales_payment_method_pub == 'Card') {
            echo __(' / Card', 'foosales');
        }
        if($foosales_payment_method =='Cash' or $foosales_payment_method_pub == 'Cash' ) {
            echo __('/ Cash', 'foosales');
        } 
        echo "</em></p>";
    }   
    
    /**
     * Adds the WooCommerce sub menu
     * 
     */
    public function add_woocommerce_submenu() {

        add_submenu_page( 'null',__( 'FooSales Introduction', 'foosales' ), __( 'FooSales Introduction', 'foosales' ), 'manage_options', 'foosales-introduction', array($this, 'add_woocommerce_submenu_page') ); 

    }

    /**
     * Adds the WooCommerce sub menu page
     * 
     */
    public function add_woocommerce_submenu_page() {
        
        require($this->Config->templatePath.'pluginintroduction.php');

    }

}

$FooSales = new FooSales();

function foosales_check_plugin_active( $plugin ) {

    return in_array( $plugin, (array) get_option( 'active_plugins', array() ) );
}